""" Primitive data type class. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: primitive_data_type_class.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging
from typing import Any
from PySide6.QtWidgets import QBoxLayout, QHBoxLayout, QVBoxLayout
from pasta_eln.dataverse.utils import is_date_time_type
from pasta_eln.GUI.dataverse.data_type_class_base import DataTypeClass
from pasta_eln.GUI.dataverse.data_type_class_context import DataTypeClassContext
from pasta_eln.GUI.dataverse.utility_functions import (add_clear_button, add_delete_button, create_date_time_widget,
                                                       create_line_edit, get_primitive_line_edit_text_value)


class PrimitiveDataTypeClass(DataTypeClass):
  """
  Represents the primitive data type class.
  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the PrimitiveDataTypeClass class.

    Explanation:
        This method creates a new instance of the PrimitiveDataTypeClass class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the PrimitiveDataTypeClass class.

    """
    return super().__new__(cls)

  def __init__(self, context: DataTypeClassContext) -> None:
    """
    Initializes a new instance of the PrimitiveDataTypeClass.

    Args:
        context (DataTypeClassContext): The context for the PrimitiveDataTypeClass.

    Explanation:
        This method initializes a new instance of the PrimitiveDataTypeClass with the provided context.
    """
    super().__init__(context)
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

  def add_new_entry(self) -> None:
    """
    Adds a new entry to the primitive data type class.

    Explanation:
        This method attempts to find the primitiveVerticalLayout in the main vertical layout.
        If the layout is found, it populates the horizontal layout with the provided type information.
        If the layout is not found, it logs an error message.

    """
    new_primitive_entry_layout = self.context.main_vertical_layout.findChild(QVBoxLayout,
                                                                             'primitiveVerticalLayout')
    if not isinstance(new_primitive_entry_layout, QVBoxLayout):
      self.logger.error('Failed to find primitiveVerticalLayout!')
      return
    value_template = self.context.meta_field.get('valueTemplate')
    first_value = str(value_template[0]) if value_template else ''
    self.populate_primitive_horizontal_layout(
      new_primitive_entry_layout,
      self.context.meta_field.get('typeName', ''),
      '',
      first_value)

  def populate_entry(self) -> None:
    """
    Populates the entry based on the meta_field information.

    Explanation:
        This method populates the entry based on the meta_field information.
        It disables the add push button if the meta_field does not have multiple entries.
        It then calls the populate_primitive_entry method to populate the entry.
    """
    if not self.context.meta_field.get('multiple'):
      self.context.add_push_button.setDisabled(True)
    self.populate_primitive_entry()

  def save_modifications(self) -> None:
    """
    Saves modifications to the meta_field based on the type and class.

    Explanation:
        This method saves changes to the meta_field based on the type and class information.
        It retrieves the primitiveVerticalLayout and updates the meta_field value accordingly.
    """
    self.logger.info('Saving changes to meta_field for type, name: %s, class: %s',
                     self.context.meta_field.get('typeName'),
                     self.context.meta_field.get('typeClass'))
    primitive_vertical_layout = self.context.main_vertical_layout.findChild(QVBoxLayout,
                                                                            'primitiveVerticalLayout')
    if primitive_vertical_layout is None:
      return
    if isinstance(primitive_vertical_layout, QVBoxLayout):
      if self.context.meta_field.get('multiple'):
        self.context.meta_field['value'].clear()
        for widget_pos in range(primitive_vertical_layout.count()):
          text = get_primitive_line_edit_text_value(primitive_vertical_layout, widget_pos)
          if text and text != 'No Value':
            self.context.meta_field['value'].append(text)
      else:
        text = get_primitive_line_edit_text_value(primitive_vertical_layout, 0)
        self.context.meta_field['value'] = text if text != 'No Value' else ''
    else:
      self.logger.error('Failed to find primitiveVerticalLayout!')

  def populate_primitive_entry(self) -> None:
    """
    Populates the primitive entry layout based on the meta_field information.

    Explanation:
        This method populates the primitive entry layout based on the meta_field information.
        It creates a new QVBoxLayout and adds widgets to it based on the meta_field values.
        The widgets are created using the create_primitive_widget function.
        The new QVBoxLayout is then added to the mainVerticalLayout of the class instance.

    """
    type_value = self.context.meta_field.get('value')
    type_name = self.context.meta_field.get('typeName', '')
    value_template = self.context.meta_field.get('valueTemplate')

    self.logger.info('Populating new entry of type primitive, name: %s', type_name)

    new_primitive_entry_layout = QVBoxLayout()
    new_primitive_entry_layout.setObjectName('primitiveVerticalLayout')
    if self.context.meta_field.get('multiple') and isinstance(type_value, list):
      template_first = value_template[0] if isinstance(value_template, list) else ''
      if not type_value:
        self.populate_primitive_horizontal_layout(
          new_primitive_entry_layout,
          type_name,
          '',
          template_first)
      else:
        for value in type_value:
          self.populate_primitive_horizontal_layout(
            new_primitive_entry_layout,
            type_name,
            value,
            template_first)
    else:
      self.populate_primitive_horizontal_layout(
        new_primitive_entry_layout,
        type_name,
        type_value or '',
        value_template if isinstance(value_template, str) else '',
        False)
    self.context.main_vertical_layout.addLayout(new_primitive_entry_layout)

  def populate_primitive_horizontal_layout(self,
                                           new_primitive_entry_layout: QBoxLayout,
                                           type_name: str,
                                           type_value: str,
                                           type_value_template: str,
                                           enable_delete_button: bool = True) -> None:
    """
    Populates the horizontal layout for a primitive entry.

    Args:
        enable_delete_button (bool): Enable or disable the delete button.
        new_primitive_entry_layout (QBoxLayout): The layout to populate.
        type_name (str): The name of the primitive type.
        type_value (str): The value of the primitive type.
        type_value_template (str): The template value for the primitive type.

    Explanation:
        This function populates the horizontal layout for a primitive entry.
        It creates a new QHBoxLayout and adds widgets to it based on the provided type_name, type_value, and type_value_template.
        The widgets can be either a QDateTimeEdit or a QLineEdit, depending on the type_name.
        The new QHBoxLayout is then added to the new_primitive_entry_layout, which is added to the main layout.

    """
    new_primitive_entry_horizontal_layout = QHBoxLayout()
    new_primitive_entry_horizontal_layout.setObjectName('primitiveHorizontalLayout')
    new_primitive_entry_horizontal_layout.addWidget(
      create_date_time_widget(type_name, type_value, self.context.parent_frame, type_value_template)
      if is_date_time_type(type_name) else
      create_line_edit(type_name, type_value, self.context.parent_frame, type_value_template))
    add_clear_button(self.context.parent_frame, new_primitive_entry_horizontal_layout)
    add_delete_button(self.context.parent_frame, new_primitive_entry_horizontal_layout, enable_delete_button)
    new_primitive_entry_layout.addLayout(new_primitive_entry_horizontal_layout)



""" Table Header dialog: change which columns are shown and in which order """
import json
import shutil
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Callable
import qrcode
import qtawesome as qta
import requests
from PIL.ImageQt import ImageQt
from PySide6.QtGui import QPixmap, QRegularExpressionValidator  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QFileDialog,  # pylint: disable=no-name-in-module
                               QLabel, QLineEdit, QMessageBox, QTextEdit, QVBoxLayout)
from ..elabFTWapi import ElabFTWApi
from ..fixedStringsJson import CONF_FILE_NAME
from ..guiCommunicate import Communicate
from ..guiStyle import IconButton, Label, TextButton, showMessage, widgetAndLayoutGrid


class ProjectGroup(QDialog):
  """ Table Header dialog: change which columns are shown and in which order """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[bool],None]):
    """
    Initialization

    Args:
      backend (Backend): PASTA-ELN backend
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm    = comm
    self.projectGroupTested = False
    self.callbackFinished = callbackFinished
    self.configuration = self.comm.backend.configuration
    self.emptyConfig:dict[str,Any] = {'local':{},'remote':{}}
    self.elabApi: ElabFTWApi|None = None
    self.serverPG: set[tuple[str,Any,Any,Any]] = set()

    # GUI elements
    mainL = QVBoxLayout(self)
    Label('Project group editor', 'h1', mainL)

    # LEFT SIDE: form
    _, self.formL = widgetAndLayoutGrid(mainL, spacing='m')
    self.selectGroup = QComboBox()
    self.selectGroup.addItems(self.configuration['projectGroups'].keys())
    self.selectGroup.currentTextChanged.connect(self.changeProjectGroup)
    self.formL.addWidget(self.selectGroup, 0, 0)
    self.groupTextField = QLineEdit()
    self.groupTextField.setValidator(QRegularExpressionValidator('\\w{3,}'))
    self.comboboxActive = True

    newButton = IconButton('fa5s.plus',    self, [Command.NEW], tooltip='New project group')
    self.formL.addWidget(newButton, 0, 2)
    delButton = IconButton('fa5s.trash',   self, [Command.DEL], tooltip='Delete project group')
    self.formL.addWidget(delButton, 0, 3)

    self.directoryLabel = QLabel('label')
    self.formL.addWidget(self.directoryLabel, 1, 0, 1, 2)
    row1Button = IconButton('fa5.edit',   self, [Command.CHANGE_DIR], tooltip='Edit data path')
    self.formL.addWidget(row1Button, 1, 3)

    self.addOnLabel = QLabel('addon')
    self.formL.addWidget(self.addOnLabel, 2, 0, 1, 2)
    row2Button = IconButton('fa5.edit',   self, [Command.CHANGE_ADDON], tooltip='Edit add-on path')
    self.formL.addWidget(row2Button, 2, 3)

    self.formL.addWidget(QLabel('Server address:'), 3, 0)
    self.serverLabel = QLineEdit('server')
    self.serverLabel.setPlaceholderText('Enter server address')
    self.formL.addWidget(self.serverLabel, 3, 1)
    self.row3Button = TextButton('Verify',   self, [Command.TEST_SERVER], tooltip='Check server')
    self.formL.addWidget(self.row3Button, 3, 3)

    self.formL.addWidget(QLabel('API-key:'), 4, 0)
    self.apiKeyLabel = QTextEdit()
    self.apiKeyLabel.setPlaceholderText('Enter API key')
    self.apiKeyLabel.setFixedHeight(48)
    # self.apiKeyLabel.setValidator(QRegularExpressionValidator(r"\d+-[0-9a-f]{85}"))
    self.formL.addWidget(self.apiKeyLabel, 4, 1)
    row4Button1 = IconButton('fa5s.question-circle', self,      [Command.TEST_API_HELP], tooltip='Help on obtaining API key')
    self.formL.addWidget(row4Button1, 4, 2)
    self.row4Button2 = TextButton('Verify',   self, [Command.TEST_APIKEY], tooltip='Check API-key')
    self.formL.addWidget(self.row4Button2, 4, 3)

    self.formL.addWidget(QLabel('Storage block:'), 5, 0)
    self.serverProjectGroupLabel = QComboBox()
    self.formL.addWidget(self.serverProjectGroupLabel, 5, 1)
    self.row5Button2 = TextButton('Verify',   self, [Command.TEST_SERVERPG], tooltip='Check access to storage block')
    self.formL.addWidget(self.row5Button2, 5, 3)

    # RIGHT SIDE: button and image
    qrButton = TextButton('Create QR code', self, [Command.CREATE_QRCODE])
    self.formL.addWidget(qrButton, 0, 6)
    self.image = QLabel()
    self.formL.addWidget(self.image, 1, 6, 4, 1)

    #final button box
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)

    #initialize
    if hasattr(self.comm.backend, 'configurationProjectGroup'):
      self.selectGroup.setCurrentText(self.comm.backend.configurationProjectGroup)
      self.selectGroup.currentTextChanged.emit(self.comm.backend.configurationProjectGroup)


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
      self.callbackFinished(False)
    elif 'Save' in btn.text() and not self.selectGroup.isHidden():
      # all information (excl. storage block) is already in self.configuration saved
      key      = self.selectGroup.currentText()
      config   = self.configuration['projectGroups'][key]
      if config['remote'].get('url','') and config['remote'].get('key','') and not self.projectGroupTested:
        showMessage(self, 'Error', 'Error: You have to select and test the storage block once successfully.')
        return
      if not config['local']['path']:
        showMessage(self, 'Error', 'Error: path to data directory is not set.')
        return
      if not config['addOnDir']:
        showMessage(self, 'Error', 'Error: add-on directory not set.')
        return
      # success
      choices = [i for i in self.serverPG if i[0]==self.serverProjectGroupLabel.currentText()]
      if choices and len(choices[0])==4:
        config['remote']['config'] = {'title':choices[0][0],
                                      'id':choices[0][1],
                                      'canRead':choices[0][2],
                                      'canWrite':choices[0][3]}
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as confFile:
        confFile.write(json.dumps(self.configuration, indent=2))
      self.callbackFinished(True)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if not self.comboboxActive:  #first button press after entering a new group name
      newKey = self.groupTextField.text()
      self.selectGroup.addItem(newKey)
      self.selectGroup.setCurrentText(newKey)
      self.configuration['projectGroups'][newKey] = self.emptyConfig
      self.formL.removeWidget(self.groupTextField)
      self.formL.addWidget(self.selectGroup, 0,0)
      self.groupTextField.hide()
      self.comboboxActive = True
    key = self.selectGroup.currentText()
    if not key:
      return
    config = self.configuration['projectGroups'][key]

    #cases
    if command[0] is Command.CHANGE_DIR:
      answer = QFileDialog.getExistingDirectory(self, 'Specify new data directory')
      if not answer:
        return
      if [i for i in Path(answer).iterdir() if i.name=='pastaELN.db']:
        button = QMessageBox.question(self, 'Question', 'Do you want to use existing PASTA ELN data?',
                                      QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
        if button == QMessageBox.StandardButton.No:
          return
      elif list(Path(answer).iterdir()):
        button = QMessageBox.question(self, 'Question', 'Do you want to use folder, which is not empty? This is not recommended.',
                                      QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
        if button == QMessageBox.StandardButton.No:
          return
      config['local']['path'] = answer
      self.directoryLabel.setText(f'Data directory: {answer}')

    elif command[0] is Command.CHANGE_ADDON:
      answer = QFileDialog.getExistingDirectory(self, 'Specify new add-on directory')
      if not answer:
        return
      button = QMessageBox.question(self, 'Question', 'Do you want to copy the add-ons from the old directory (recommended)?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.Yes:
        shutil.copytree(config['addOnDir'], answer, dirs_exist_ok=True)
      config['addOnDir'] = answer
      self.addOnLabel.setText('Add on directory: ' + config['addOnDir'])

    elif command[0] is Command.TEST_SERVER:
      headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
      url     = self.serverLabel.text().strip()
      if not url:
        return
      if not url.endswith('/'):
        url += '/'
      if not url.endswith('api/v2/'):
        url += 'api/v2/'
      if not url.startswith('http'):
        url = f'https://{url}'
      config['remote']['url'] = url
      self.serverLabel.setText(url)
      try:
        requests.get(f'{url}info', headers=headers, verify=True, timeout=15)
        self.changeButtonOnTest(True, self.row3Button)
      except Exception:
        self.changeButtonOnTest(False, self.row3Button, 'Wrong server address')

    elif command[0] is Command.TEST_API_HELP:
      link = f'OR go to {config["remote"]["url"][:-7]}ucp.php?tab=4\n\n' if config['remote'].get('url','') else ''
      showMessage(self, 'Help', '### How to get an api key to access the server:\nOn the elabFTW server:\n1. go to the USER SYMBOL\n2. User-(Control) panel\n3. API KEYS\n\n'\
                  f'{link}'\
                  '1. Specify a name: e.g. "pasta_eln"\n2. change the permissions to "Read/Write"\n3. click "Generate new API key"\n\nCopy-paste that key into the text box on the right-hand-side')

    elif command[0] is Command.TEST_APIKEY:
      if self.apiKeyLabel.toPlainText()!='--- API key hidden ---':
        config['remote']['key'] = self.apiKeyLabel.toPlainText().strip()
      headers = {'Content-type': 'application/json', 'Accept': 'text/plain', 'Authorization':config['remote']['key']}
      url = config['remote']['url']
      try:
        res = requests.get(f'{url}info', headers=headers, verify=True, timeout=15)
        if res.status_code==200:
          elabVersion = int(json.loads(res.content.decode('utf-8')).get('elabftw_version','0.0.0').split('.')[0])
          if elabVersion<5:
            showMessage(self, 'Error', 'Old elabFTW server installation')
          # success
          self.changeButtonOnTest(True, self.row4Button2)
          self.elabApi   = ElabFTWApi(url, config['remote']['key'])
          response = self.elabApi.readEntry('items?q=category%3AProjectGroup&archived=on')
          if not response:
            showMessage(self, 'Error', 'Please ask your database admin to add your project-group(s).')
          self.serverPG = {(i['title'],i['id'],i['canread'],i['canwrite']) for i in response}
          self.serverProjectGroupLabel.clear()
          self.serverProjectGroupLabel.addItems([i[0] for i in self.serverPG])
        else:
          self.changeButtonOnTest(False, self.row4Button2, 'Wrong API key')
      except Exception:
        self.changeButtonOnTest(False, self.row4Button2, 'Wrong API key')

    elif command[0] is Command.TEST_SERVERPG and self.elabApi is not None:
      idxList = [i[1] for i in self.serverPG if i[0]==self.serverProjectGroupLabel.currentText()]
      if not idxList:
        return
      idx = idxList[0]
      currentBody = self.elabApi.readEntry('items',idx)[0]['body']
      currentBody+= f'<br>Tested access by {self.configuration["userID"]} on {datetime.now().isoformat()} <br>'
      if self.elabApi.updateEntry('items',idx, {'body':currentBody}):
        self.projectGroupTested = True
        self.changeButtonOnTest(True, self.row5Button2)
      else:
        self.changeButtonOnTest(False, self.row5Button2, 'You do not have access to this project group')

    elif command[0] is Command.CREATE_QRCODE:
      text   = json.dumps(config['remote'])
      img    = qrcode.make(text, error_correction=qrcode.constants.ERROR_CORRECT_M).get_image()
      pixmap = QPixmap.fromImage(ImageQt(img).scaledToWidth(350))
      self.image.setPixmap(pixmap)

    elif command[0] is Command.NEW:
      self.formL.removeWidget(self.selectGroup)
      self.formL.addWidget(self.groupTextField, 0, 0)
      self.directoryLabel.setText('Data directory: ')
      self.addOnLabel.setText('Add on directory: ')
      self.serverLabel.setText('')
      self.apiKeyLabel.setText('')
      self.image.setPixmap(QPixmap())
      self.comboboxActive = False

    elif command[0] is Command.DEL:
      del self.configuration['projectGroups'][key]
      self.selectGroup.removeItem(self.selectGroup.currentIndex())
      self.selectGroup.setCurrentIndex(0)

    else:
      print('Got some button, without definition', command)
    return


  def changeProjectGroup(self, item:str) -> None:
    """
    change the project group to this; do not save to file

    Args:
      item (str): name of project group
    """
    config = self.configuration['projectGroups'].get(item, self.emptyConfig)
    self.directoryLabel.setText('Data directory: ' + config['local'].get('path',''))
    self.addOnLabel.setText('Add on directory: ' + config.get('addOnDir',''))
    self.serverLabel.setText(config['remote'].get('url', ''))
    if config['remote'].get('key', ''):
      self.apiKeyLabel.setText('--- API key hidden ---')
    else:
      self.apiKeyLabel.setText('')
    return


  def changeButtonOnTest(self, success:bool, button:TextButton, message:str='') -> None:
    """ Helper function to change buttons upon success/failure

    Args:
      success (bool): change to successful state
      button (TextBotton): button to change
      message (str): optional message on failure
    """
    if success:
      button.setStyleSheet('background: #00FF00')
      button.setText('')
      button.setIcon(qta.icon('fa5s.check-square', scale_factor=1))
    else:
      if message:
        showMessage(self, 'Error', message)
      button.setStyleSheet('background: #FF0000')
      button.setText('')
      button.setIcon(qta.icon('fa5.times-circle', scale_factor=1))
    return


class Command(Enum):
  """ Commands used in this file """
  NEW          = 1
  DEL          = 2
  CHANGE_DIR   = 3
  CHANGE_ADDON = 4
  TEST_SERVER  = 5
  TEST_APIKEY  = 6
  TEST_API_HELP= 7
  CREATE_QRCODE= 8
  TEST_SERVERPG= 9



""" Represents the controlled vocabulary data type class. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: controlled_vocabulary_data_type_class.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from typing import Any
from PySide6.QtWidgets import QComboBox, QHBoxLayout
from pasta_eln.GUI.dataverse.data_type_class_base import DataTypeClass
from pasta_eln.GUI.dataverse.data_type_class_context import DataTypeClassContext
from pasta_eln.GUI.dataverse.utility_functions import add_clear_button, add_delete_button


class ControlledVocabularyDataTypeClass(DataTypeClass):
  """
  Represents the controlled vocabulary data type class.
  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the ControlledVocabularyDataTypeClass class.

    Explanation:
        This method creates a new instance of the ControlledVocabularyDataTypeClass class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the ControlledVocabularyDataTypeClass class.

    """
    return super().__new__(cls)

  def __init__(self, context: DataTypeClassContext) -> None:
    """
    Initializes a new instance of the ControlledVocabularyDataTypeClass class.

    Explanation:
        This method initializes a new instance of the ControlledVocabularyDataTypeClass class.
    Args:
      context (DataTypeClassContext): The context of the data type class.
    """
    super().__init__(context)
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

  def add_new_entry(self) -> None:
    """
    Adds a new entry to the controlled vocabulary.

    Explanation:
        This method adds a new entry to the controlled vocabulary.
    """
    value_template = self.context.meta_field.get('valueTemplate')
    if self.context.meta_field.get('multiple'):
      self.add_new_vocab_entry(value_template, value_template[0] if value_template else None)
    else:
      self.add_new_vocab_entry(
        ['No Value', value_template] if value_template else ['No Value'],
        value_template or 'No Value'
      )

  def populate_entry(self) -> None:
    """
    Populates the entry with the current value of the controlled vocabulary.

    Explanation:
        This method populates the entry with the current value of the controlled vocabulary.
    """
    value_block = self.context.meta_field.get('value')
    value_template = self.context.meta_field.get('valueTemplate')
    if self.context.meta_field.get('multiple'):
      if value_block:
        for value in value_block:
          self.add_new_vocab_entry(value_template, value)
      else:
        self.add_new_vocab_entry(value_template, None)
    else:
      self.add_new_vocab_entry(
        ['No Value', value_template] if value_template else ['No Value'],
        value_block or 'No Value'
      )

  def save_modifications(self) -> None:
    """
    Saves the modifications to the controlled vocabulary.

    Explanation:
        This method saves the modifications to the controlled vocabulary.
    """
    if self.context.meta_field.get('multiple'):
      self.save_multiple_entries()
      self.logger.info('Saved multiple entry modifications successfully, value: %s', self.context.meta_field)
    elif layout := self.context.main_vertical_layout.findChild(QHBoxLayout,
                                                               'vocabHorizontalLayout'):
      if not isinstance(layout, QHBoxLayout):
        self.logger.error('vocabHorizontalLayout not found!')
        return
      self.save_single_entry(layout)
      self.logger.info('Saved single entry modification successfully, value: %s', self.context.meta_field)
    else:
      self.logger.warning('Failed to save modifications, no layout found.')

  def save_single_entry(self, layout: QHBoxLayout) -> None:
    """
    Saves the single entry to the controlled vocabulary.

    Explanation:
        This method saves the single entry to the controlled vocabulary.
    Args:
      layout (QHBoxLayout): The layout of the single entry.
    """
    if combo_box := layout.itemAt(0).widget():
      if not isinstance(combo_box, QComboBox):
        self.logger.error('Combo box not found!')
        return
      current_text = combo_box.currentText()
      self.context.meta_field['value'] = current_text if current_text and current_text != 'No Value' else None
    else:
      self.context.meta_field['value'] = None
      self.logger.error('Combo box not found!')

  def save_multiple_entries(self) -> None:
    """
    Saves the multiple entries to the controlled vocabulary.

    Explanation:
        This method saves the multiple entries to the controlled vocabulary.
    """
    value = self.context.meta_field.get('value')
    if not isinstance(value, list):
      self.logger.error('Value is not a list')
      return
    value.clear()
    for layout_pos in reversed(range(self.context.main_vertical_layout.count())):
      if vocab_horizontal_layout := self.context.main_vertical_layout.itemAt(layout_pos).layout():
        combo_box = vocab_horizontal_layout.itemAt(0)
        if combo_box is not None:
          combo_boxW = combo_box.widget()
          if not isinstance(combo_boxW, QComboBox):
            continue
          text = combo_boxW.currentText()
          if text and text != 'No Value':
            value.append(text)
    self.context.meta_field['value'] = list(set(value))

  def add_new_vocab_entry(self,
                          controlled_vocabulary: list[str] | None = None,
                          value: str | None = None) -> None:
    """
    Adds a new vocabulary entry to the controlled vocabulary frame.

    Args:
        controlled_vocabulary (list[str] | None): The list of controlled vocabulary options.
        value (str | None): The initial value for the vocabulary entry.

    Explanation:
        This method adds a new vocabulary entry to the controlled vocabulary frame.
        It creates a new layout with a combo box and a delete button, and adds it to the main vertical layout.
        The combo box is populated with the controlled vocabulary options, and the initial value is set if provided.
        The delete button is connected to the delete_layout_and_contents function to handle deletion of the entry.
    """
    new_vocab_entry_layout = QHBoxLayout()
    new_vocab_entry_layout.setObjectName('vocabHorizontalLayout')
    combo_box = QComboBox(parent=self.context.parent_frame)
    combo_box.setObjectName('vocabComboBox')
    combo_box.setToolTip('Select the controlled vocabulary.')
    combo_box.addItems(controlled_vocabulary or [])
    combo_box.setCurrentText(value or '')
    new_vocab_entry_layout.addWidget(combo_box)
    add_clear_button(self.context.parent_frame, new_vocab_entry_layout)
    add_delete_button(self.context.parent_frame, new_vocab_entry_layout)
    self.context.main_vertical_layout.addLayout(new_vocab_entry_layout)
